//walker.js
const fs = require('fs'),
crypto = require('crypto'),
path = require('path'),
{EventEmitter} = require('events');

module.exports = class FileWalker extends EventEmitter {

constructor (debug){
 super();
 this.debug = debug ? true : false;
 this.reset(debug);
}
reset(){
 this.isPaused = false;
 this.queue = [];
 this.files = {};
 this.dupFiles = {};
 this.filter_dir = () => false;
 this.filter_file = () => false;
}
addToQueue(entry){
 Array.isArray(entry) 
  ? Array.prototype.push.apply (this.queue, entry)
  : this.queue.push (entry)
}
filterDir (fn){
 this.filter_dir = fn;
}
filterFile (fn){
 this.filter_file = fn;
}
start(){
 this.emit('start');
 this.debug&&console.log('walker.start')
 this.next();
}
fetch (entry) {
 fs.lstat(entry, (err,stat) => {
  if (err){
   this.debug&&console.log('walker.fetch Error stat: '+entry);
   this.emit('error',err,entry,stat);
   return this.next();
  }

  if (stat.isFile()){
   if (this.filter_file(entry,stat)){
    this.debug&&console.log('walker.fetch filterFile: '+entry);
    return this.next();
   }
   this.debug&&console.log('walker.fetch File: '+entry);
   this.emit('file',entry,stat);
   this.generateHash(entry,stat);/*Read file header*/
   this.next();
  }
  else if (stat.isDirectory()){
   if (this.filter_dir(entry,stat)){
    this.debug&&console.log('walker.fetch filterDir: '+entry);
    return this.next();
   }
   this.debug&&console.log('walker.fetch Dir: '+entry);
   this.emit('dir',entry,stat);
   fs.readdir(entry, (err,files) => {
    if (err){
     this.debug&&console.log('walker.fetch Error readdir: '+entry);
     this.emit('error',err,entry,stat);
     return this.next();
    }
    Array.prototype.push.apply(this.queue, files.map( file => {
     return path.join(entry,file);
    }));
	this.next();
   });
  }
  else {
   this.debug&&console.log('walker.fetch unknown  or inaccessilbe: '+entry);
   this.emit('unknown',entry,stat);
   this.next()
 }

 });
}

next(){
 if (this.isPaused) {
  this.emit('pause');
  this.debug&&console.log('walker.next isPaused, remaining files in queue:'+this.queue.length);
  return this;
 }
 let nextEntry = this.queue.shift();
 if (!nextEntry){
  this.emit('done');
  this.debug&&console.log('walker.next Done');
  return this;
 }
 this.fetch(nextEntry);
}

pause(){
 if (this.isPaused === true){
  this.debug&&console.log('walker.pause isPaused failed. App was already paused');
  return this;	 
 }
 this.isPaused = true;
}

resume(){
 if (this.isPaused === false){
  this.debug&&console.log('walker.resume Resume failed. App was already resumed');
  return;	 
 }
 this.isPaused = false;
 this.debug&&console.log('walker.resume Resume');
 this.emit('resume');
 this.next();
}

generateHash(file,stat){
 const defaultLength = 4200,
 len = stat.size < defaultLength ? stat.size : defaultLength,
 pos = 0, offset =0;

 fs.open(file, 'r', (err, fd) => {
  if (err) {
   this.emit('error',err,file,stat);
   this.debug&&console.log('walker.generatehash fs.open error: '+err);
   return;
  }

  const buffer = Buffer.alloc(len);
  fs.read(fd, buffer, offset, len, pos, (err, bytesRead, buffer) => {
   if (err){ 
    this.emit('error',err,file,stat);
    this.debug&&console.log('walker.generateHash fs.read.error'+err);
    return;
   }
   fs.close(fd, (err) => {
    if (err){
     this.emit('error',err,file,stat);
     this.debug&&console.log('walker.generateHash fs.close.error'+err);
     return;
    }
   });
   
   const hash = crypto
             .createHash('whirlpool')
             .update(buffer)
             .digest('hex');
   
   this.emit('hash',file,stat,hash);
   this.debug&&console.log('walker.generateHash hash emitted');
   this.findDuplicate(file,stat,hash);
   return;
  });
 })
}

findDuplicate(file,stat,hash){
 const ext  = path.extname(file),
 size = stat.size;
 
 let hashExist = this.files[hash]
               ? this.files[hash]
              : (this.files[hash] = {}),
 
 extExist = hashExist[ext]
          ? hashExist[ext]
         : (hashExist[ext] = {}),

 sizeExist = extExist[size];
 
 if (sizeExist === undefined){
  extExist[size] = file;
  return;
 }
 
let hashDExist = this.dupFiles[hash]
               ? this.dupFiles[hash]
               : (this.dupFiles[hash] = {}),

 extDExist = hashDExist[ext]
             ? hashDExist[ext]
             : (hashDExist[ext] = {}),
             
 sizeDExist = extDExist[size];
 
 if (sizeDExist === undefined){
  let duplicates = [sizeExist,file];
  extDExist[size] = duplicates;
  this.emit('duplicate',duplicates,size,ext,hash);
  this.debug&&console.log('walker.findDuplicate duplicatesArray emitted');
  return;
 }
 
 sizeDExist.push(file);
 this.emit('duplicate',[file],size,ext,hash,sizeDExist);
 this.debug&&console.log('duplicateSingle emitted');
 return;
}
}
